
/**
 * Simulation of EG8010 output signals.
 * Based on Poida's inverter code. 
 */


#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))


#define NPWM          200       // Half sine points
#define PPWM          800       // Timer1 TOP value to achieve 20KHz
#define RAMPUP_POWER  0.01      // ramp up power step

uint8_t sinPositive, outEnable;
uint16_t sinIdx;
uint32_t vpwr, sst;
uint16_t l[NPWM + 1];
float pwr, vSet, vFb;
uint16_t dbounce;

double Input, Output, Setpoint;
double errSum, lastErr, error, dErr;
double kp, ki, kd;

float xv[5], yv[5];

void initVars()
{
    sinPositive = 1;
    sst = 0;
    sinIdx = 0;
    vSet = vFb = 0.0;
    pwr = 0.0;
    vpwr = 0;
    errSum = 0.0;
}

void setup()
{
    // can NOT let arduino system alter any pins from D0 to D7 via calls to pinMode() etc.. It buggers d5 which I need to pull down low-Z upon boot
    DDRD = 0xFF; 
    cbi(PORTD, 5);    // disable MOSFET drivers
    
    // generate half sine table
    float t = 0.0;
    for (int i = 0; i <= NPWM; i++)
    {
        l[i] = (int)(16384.0 * sin(t));
        t += 3.14159 / (float)NPWM;
    }

    initVars();

    outEnable = 0;
    dbounce = 0;

    // set up PID 0.1 0.0 0.01 for biquad LP Fc = 0.02, np=4
    kp = 0.1;
    ki = 0.0;
    kd = 0.01;

    pinMode(12, OUTPUT);        // PB4 - 50Hz  SPWMOUT1
    pinMode(13, OUTPUT);        // PB5 - 50Hz  SPWMOUT2
    pinMode(9, OUTPUT);         // PB1 - 20KHz SPWMOUT3
    pinMode(10, OUTPUT);        // PB2 - 20KHz SPWMOUT4
    pinMode(8, INPUT_PULLUP);   // Start/Stop button
    
    noInterrupts();
    
    // Timer1 configured for 50us (20KHz) period; fast PWM mode 14; channel A/B complementary output; interrupt on overflow
    TCCR1A = _BV(COM1A1) | _BV(COM1B1) | _BV(COM1B0) | _BV(WGM11);
    TCCR1B = _BV(WGM13) | _BV(WGM12) | _BV(CS10);
    OCR1A = 0;
    OCR1B = 0;
    ICR1 = PPWM;
    TIMSK1 |= (1 << TOIE1);

    interrupts(); // enable all interrupts
    TIMSK0 = 0;   // except arduino clock

    sbi(ADCSRA, ADPS2); // ADC clock prescale = 16 now. this saves 100us in main loop with little cost to loop stability
    cbi(ADCSRA, ADPS1);
    cbi(ADCSRA, ADPS0);

    Serial.begin(115200);
}

// Timer1 overflow interrupt at 20Khz
ISR(TIMER1_OVF_vect)
{ 
    if (sinIdx == 0)
    {
        if (sinPositive)
        {
            // sync pulse for DSO, alligned with start of 50Hz
            sbi(PORTD, 7); 
            cbi(PORTD, 7);
            
            // toggle complementary 50Hz outputs
            PORTB = (PORTB & ~(_BV(4) | _BV(5))) | _BV(5);
        }
        else 
        {
            // toggle complementary 50Hz outputs
            PORTB = (PORTB & ~(_BV(4) | _BV(5))) | _BV(4);
        }
        sinPositive = !sinPositive;
    }

    // next sine step
    uint32_t c = (l[sinIdx] * vpwr) >> 14;

    if (sinPositive)
    {
        OCR1A = c;
        OCR1B = c;
    }
    else
    {
        uint16_t ocr = PPWM - c;
        OCR1A = ocr;
        OCR1B = ocr;
    }
    if (++sinIdx >= NPWM)
    {
      sinIdx = 0;
    }
}

void do_pid()
{
    double timeChange = 0.01; // always called at 100Hz, so dT will be 10msec..
    error = Setpoint - Input;
    errSum += (error * timeChange);
    dErr = (error - lastErr) / timeChange;
    Output = kp * error + ki * errSum + kd * dErr;
    lastErr = error;
}

void loop()
{
    checkSwitch();

    // filter vSet,0 analog input and scale to 0 to 1
    // vFb = output ACV, via transformer, 220VAC = 3.5V DC
    // vSet = setpoint. About 4.3kHz sample rate, or 230us per 2ch sample and filter plus 2 x output to SPI DAC
    float tc2 = 0.01;
    vSet = (1.0 - tc2) * vSet + tc2 * analogRead(2) / 1024.0;

    float fvFb = analogRead(0) / 1024.0;
    // this is good, bessel filter of some Fc.
    xv[0] = xv[1];
    xv[1] = xv[2];
    xv[2] = xv[3];
    xv[3] = xv[4];
    xv[4] = fvFb / 2.259747797e+05;
    yv[0] = yv[1];
    yv[1] = yv[2];
    yv[2] = yv[3];
    yv[3] = yv[4];
    yv[4] = (xv[0] + xv[4]) + 4 * (xv[1] + xv[3]) + 6 * xv[2] + (-0.7428578687 * yv[0]) + (3.1954036268 * yv[1]) + (-5.1599230905 * yv[2]) + (3.7073065280 * yv[3]);
    vFb = yv[4];

    if (outEnable)
    {
        if (sinIdx == 0)
        {
            // slow start timer counts to 250 with gentle ramp up, then switches over to PID control. (gulp!)
            if (sst < 250) 
            {
                // comment next line to prevent PID control, it will remain in simple bang-bang control
                sst++; 
                if (vFb < vSet)
                {
                    pwr += RAMPUP_POWER;
                }
                else
                {
                    pwr -= RAMPUP_POWER;
                }
            }
            else
            {
                Setpoint = vSet;
                Input = vFb;
                do_pid();
                pwr = pwr + Output;
            }
            if (pwr > 1.0)
            {
                pwr = 1.0;
            }
            if (pwr < 0.01)
            {
                pwr = 0.01;
            }
            vpwr = (int)((float)PPWM * pwr);
        }
    }
}

void checkSwitch()
{
    static uint8_t btnDown = 0;
  
    dbounce++;
    if (digitalRead(8) == 0) // momentary button pressed, pulling internal pullup low
    {
        if (!btnDown) 
        {
            btnDown = 1;  
            if (outEnable == 0)
            { 
                // start
                //Serial.println("Start");
                outEnable = 1;
                sbi(PORTD, 5);    // enable MOSFET drivers
            }
            else
            { 
                // stop 
                //Serial.println("Stop");
                outEnable = 0;
                cbi(PORTD, 5);    // disable MOSFET drivers
            }
            initVars();
            dbounce = 0;
        }
    }
    else 
    {
        if (dbounce > 100)
        {
            btnDown = 0;
        }
        
    }
}
